############################################################################
# miLD -- program for calculation of different LD measures
# (C) 2002 Y. Aulchenko, EMC Rotterdam, Inst. Cytol. Genet. SD RAS
#
#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
#     
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#     GNU General Public License for more details.
#     
#     You should have received a copy of the GNU General Public License
#     along with this program; if not, write to the Free Software
#     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# and now it is version 2.1
# modified 03-09-2002 (rotate.f now calculates bootstrap, account 
# for it in miLD...
# modified 04-09-2002 -- comand line for additional options
# modified 11-11-2002 -- Dboo output added
#
# contact: yuriisa@bionet.nsc.ru
############################################################################


############################################################################
# DOS. Uncomment the following section for DOS platform and comment for UNIX
#	$roprog = './rotate.exe';
#	$mldprog = './mld.exe';
############################################################################
# UNIX. Uncomment the following section for Unix-like platform and comment for DOS
	$roprog = './rotate';
	$mldprog = './mld';
############################################################################

$data_file="haplotype.data";
$output_file="results.csv";
$pairs_mode=1;
$rotate_mode=1;
$scale_factor=1.0;
$quant1=0.025;
$quant2=0.975;
$dump_boo=0;

#foreach my $arg (@ARGV) {
#    $data_file=$1 if ($arg =~ /^-d(.*)$/);
#    }
#$line = join (' ',@ARGV);
for ($i=0;$i<@ARGV;$i++) {

    $line = $ARGV[$i];

    $data_file=$1 if ($line =~ /^-d *(.[^\s]*)/);
    $output_file=$1 if ($line =~ /^-o *(.[^\s]*)/);
    $pairs_mode=$1 if ($line =~ /^-e *(.[^\s]*)/);
    $scale_factor=$1 if ($line =~ /^-s *(.[^\s]*)/);
    $quant1=$1 if ($line =~ /^-q1 *(.[^\s]*)/);
    $quant2=$1 if ($line =~ /^-q2 *(.[^\s]*)/);
    $rotate_mode=$1 if ($line =~ /^-bm *(.[^\s]*)/);
    $dump_boo=$1 if ($line =~ /^-bd *(.[^\s]*)/);

    $data_file=$ARGV[$i+1] if ($line =~ /^-d$/);
    $output_file=$ARGV[$i+1] if ($line =~ /^-o$/);
    $pairs_mode=$ARGV[$i+1] if ($line =~ /^-e$/);
    $scale_factor=$ARGV[$i+1] if ($line =~ /^-s$/);
    $quant1=$ARGV[$i+1] if ($line =~ /^-q1$/);
    $quant2=$ARGV[$i+1] if ($line =~ /^-q2$/);
    $rotate_mode=$ARGV[$i+1] if ($line =~ /^-bm$/);
    $dump_boo=$ARGV[$i+1] if ($line =~ /^-bd$/);

}

print "
miLD version 2.1\n";

print "
Options
\t-d(haplotype-file)  : provide the haplotype file name
\t-o(output-file)     : provide the output file name
\t-e(ext-number)      : extension-number -- use an integer (how many 
\t                           consequtive markers for pairvise LD) or 
\t                          \"all\" (all pairwise LDs)
\t-bm(bootstrap-mode) : specify bootstrap-mode (0 no boo, 1 boo for D', 
\t                           2 boo for D' and D'adj)
\t-s(scale-factor)    : a real number for scale (used in bootstrap-
\t                           option 2)
\t-q1(quantile 1)     : a real number -- 1st quantile for bootstrap
\t-q2(quantile 1)     : a real number -- 2nd quantile for bootstrap
\t-bd(boot-dump)      : 0 (do not dump) or 1 (dump the bootstrap-
\t                           distribution)\n
";

print "
The following options are in effect:
\tHaplotype file          : $data_file
\tOutput file             : $output_file
\tExtend calculations     : $pairs_mode
\tBootstrap-mode          : $rotate_mode";

if ($rotate_mode ne "0") {
print "
\tScale-factor            : $scale_factor" if ($rotate_mode eq "2");
print "
\tQuantiles for bootstrap : $quant1 and $quant2
\tDump bootstrap          : $dump_boo\n\n";
}

# remove seed-file
# `rm seed.rlux`;

$time=localtime();
$time0 = time();
print"Run started $time\n\n";

if ($pairs_mode ne "all"){
die "wrong ext-number (use integer or \"all\")\n" if ($pairs_mode =~ /[^0-9]/);
}
die "wrong bootstrap-mode (use 0, 1 or 2)\n" if ($rotate_mode ne "0" and $rotate_mode ne "1" and $rotate_mode ne "2");
die "wrong bootstrap dump mode (use 0 or 1)\n" if ($dump_boo ne "0" and $dump_boo ne "1");
die "wrong scale factor (use integer or real)\n" if ($scale_factor =~ /[^0-9.]/);
die "wrong quantile 1 (use integer or real)\n" if ($quant1 =~ /[^0-9.]/);
die "wrong quantile 2 (use integer or real)\n" if ($quant2 =~ /[^0-9.]/);

# initiate dump-files (ordered bootstraped D's)
if ($rotate_mode != 0 and $dump_boo == 1) {
open (TMP,">booDprim.dump");
print TMP "Dump-file for Dprim (miLD run started at $time)\n";
close TMP;
if ($rotate_mode == 2) {
open (TMP,">booDadj.dump");
print TMP "Dump-file for Dadj (miLD run started at $time)\n";
close TMP;
}
}

open (HAP,"$data_file") or die "cannot open hapfile $data_file: $!\n";
print "reading haplotypes... ";
$len = -10;
$nchr=0;
while (<HAP>) {
	chomp;
	next if (/^$/ or /^ +$/ or /^#/);
	@line = split /\s+/;
	$len = @line if ($len == -10);
	die "different number of markers: data line $nchr+1\n" if (@line != $len);
	$srcdat[$nchr++] = $_;
}
$nmrk = $len - 4;
print "$nmrk markers in $nchr chromosomes. Done\n\n";
close (HAP);
$pairs_mode=$nmrk if ($pairs_mode eq "all");

open (OUF,">$output_file") or die "cannot open file $output_file for output: $!\n";

# calculate number of LDs to evaluate
$fstrun=0;
$enum = 0;
for ($i=1;$i<$nmrk;$i++) { 
if ($i+$pairs_mode<$nmrk) {$nlim=$i+$pairs_mode} else {$nlim=$nmrk}; 
for ($j=$i+1;$j<=$nlim;$j++) {
	$enum++;
}
}

#print "  m1  m2 nm1 nm2 nmb    Dpij   Dprim    Pval  Dstoch  VarDsto     Dadj  na1  na2  ngen       chisq\n";
#print OUF "m1,m2,nm1,nm2,nmb,Dpij,Dprim,Pval,Dstoch,VarDsto,Dadj,na1,na2,ngen,chisq,dpn,$quant1-quantile-Dprim,$quant2-quantile-Dprim,$quant1-quantile-Dadj,$quant2-quantile-Dadj\n";
print OUF "m1,m2,nm1,nm2,nmb,Dprim,Pval,Dstoch,VarDsto,Dadj,na1,na2,ngen,chisq,Dboo,$quant1-quantile-Dprim,$quant2-quantile-Dprim,$quant1-quantile-Dadj,$quant2-quantile-Dadj\n";
$cust=1;
for ($i=1;$i<$nmrk;$i++) { 
if ($i+$pairs_mode<$nmrk) {$nlim=$i+$pairs_mode} else {$nlim=$nmrk}; 
for ($j=$i+1;$j<=$nlim;$j++) {
	@ldres=&calcLD($i,$j);
	if ($ldres[0]) {
		$dadj=$intr_dprim-$meanld;
#		printf ("%4i%4i%4i%4i%4i%8.5f%8.5f%8.5f%8.5f%9.5f%9.5f%5i%5i%6i%12.5f\n",$i,$j,$znm1,$znm2,$znmb,$ldres[1],$ldres[2],$ldres[3],$meanld,$varld,$dadj,$nalm1,$nalm2,$df,$chisq);
#		print OUF "$i,$j,$znm1,$znm2,$znmb,$ldres[1],$ldres[2],$ldres[3],$meanld,$varld,$dadj,$nalm1,$nalm2,$df,$chisq,$intr_dprim,$intr_1stq,$intr_2ndq,$intr_3rdq,$intr_4thq\n";
		print OUF "$i,$j,$znm1,$znm2,$znmb,$intr_dprim,$ldres[3],$meanld,$varld,$dadj,$nalm1,$nalm2,$df,$chisq,$meanboold,$intr_1stq,$intr_2ndq,$intr_3rdq,$intr_4thq\n";
		$fstrun++;
		if ($fstrun == 1) {
			$took = time() - $time0;
			$etime = $took * $enum;
			print "First run took $took seconds. Estimated running time is $etime seconds\n";
			print "Estimated time is very innacurate (1st run-time <3 sec.)\n" if ($took<3);
			print "\nProgress:     ";
			$step = $enum / 50.0;
		} else {
			if ($enum>50) {
			if ($fstrun>=$cust*$step) {
				$proc = $cust*2;
				$proc = "0$proc" if ($proc<10);
				print "\b\b\b\bO ${proc}\%";
				$cust++;
			}} else {
				$proc = $cust;
				$proc = "0$proc" if ($proc<10);
				print "\b\b\bO ${proc}";
				$cust++;
			}
		}
	}
}}


close OUF;

$time=localtime();
$timeN=time();
$took = $timeN-$time0;
print"\n\nRun finished $time in $took seconds\n\n";

sub calcLD {
	my($nm1,$nm2) = ($_[0],$_[1]);
	my ($i,$j,$t1,$t2);
	my (@m1) = ($nchr x -10);
	my (@m2) = ($nchr x -10);
	my(@line);	
	$znm1=0;$znm2=0;$znmb=0;
	open (MLD,">temp.tmp") or die "cannot open temp.tmp for output: $!\n";
	open (MLD1,">temp.rot") or die "cannot open temp.tmp for output: $!\n";
	print MLD1 "$rotate_mode $scale_factor $quant1 $quant2 $dump_boo\n";
	for ($i=0;$i<$nchr;$i++) {
		@line = split /\s+/,$srcdat[$i];
		$m1[$i] = $line[$nm1+3];
		$m2[$i] = $line[$nm2+3];
		$t1 = $m1[$i]; $t1 = "?" if ($t1 == 0);
		$t2 = $m2[$i]; $t2 = "?" if ($t2 == 0);
		print MLD "$t1 x $t2 x\n";
		if ($m1[$i] != 0 and $m2[$i] != 0) {
			$t1 = $m1[$i]; 
			$t2 = $m2[$i]; 
			print MLD1 "$t1 $t2\n";
		}
		$znm1++ if ($m1[$i]!=0);
		$znm2++ if ($m2[$i]!=0);
		$znmb++ if ($m1[$i]!=0 and $m2[$i]!=0);
	}
	close MLD;
	print MLD1 "-999 -999\n";
	close MLD1;
	`$mldprog -paramfile=mld.par`;
	open(TMP,"mld.out") or die "cannot open mld.out: $!\n";
	while(<TMP>) {chomp;@line=split/=/;$pval=$line[1];}

	$lddd = `$roprog < temp.rot`;
	chomp $lddd; $lddd =~ s/^ +//g;
	@lilili = split /\s+/,$lddd;
	$meanld = $lilili[0];$varld=$lilili[1];
	$intr_dprim=$lilili[2];
	$intr_1stq=$lilili[3];
	$intr_2ndq=$lilili[4];
	$intr_3rdq=$lilili[5];
	$intr_4thq=$lilili[6];
	
$meanboold=$lilili[10];

	my(@trm1,@trm2);
	$n2m=0;
	for ($i=0;$i<$nchr;$i++) {
		if ($m1[$i] != 0 and $m2[$i] != 0) {
			$trm1[$n2m]=$m1[$i];
			$trm2[$n2m++]=$m2[$i];
		}
	}
#	print "number of chromosomes with alleles at both loci = $n2m\n";
#	return (0,0,0,0) if ($n2m < $nchr / 2);
	for ($i=0;$i<$n2m;$i++) {
		$cou1{$trm1[$i]}=0 if (!exists($cou1{$trm1[$i]}));
		$cou2{$trm2[$i]}=0 if (!exists($cou2{$trm2[$i]}));
		$couH{"$trm1[$i]-$trm2[$i]"}=0 if (!exists($couH{"$trm1[$i]-$trm2[$i]"}));
		$cou1{$trm1[$i]}++;
		$cou2{$trm2[$i]}++;
		$couH{"$trm1[$i]-$trm2[$i]"}++;
	}
	foreach $key (sort(keys %couH)) {
		$freqH{$key}=$couH{$key}/$n2m;
#		print "haps: $key,$couH{$key},$freqH{$key}\n";
	}
	$nalm1=0;
	foreach $key (sort(keys %cou1)) {
		$freq1{$key} = $cou1{$key}/$n2m;
		$nalm1++ if ($cou1{$key}>0);;
#		print "s: $key,$cou1{$key},$freq1{$key}\n";
	}
	$nalm2=0;
	foreach $key (sort(keys %cou2)) {
		$freq2{$key} = $cou2{$key}/$n2m;
		$nalm2++ if ($cou2{$key}>0);
#		print "s: $key,$cou2{$key},$freq2{$key}\n";
	}
	$dprim=0.0;
	$dprim1=0.0;
	$dprj=0.0;
	$chisq=0.0;
	$df=0;
	foreach $key1 (sort(keys %cou1)) {
	foreach $key2 (sort(keys %cou2)) {
		$hap = "$key1-$key2";
		$hapf =0.0; $hapf = $freqH{$hap} if (exists($freqH{$hap}));
		$pi=$freq1{$key1};$pj=$freq2{$key2};
		$pij = $pi*$pj;
		$dloc = $hapf - $pij;
		if ($dloc<0) { 
			$q1=$pij;
			$q2=(1.0-$pi)*(1.0-$pj);
			if ($q1<$q2) {$dmax=$q1;} else {$dmax=$q2;}
		} else {
			$q1=(1.0-$pi)*$pj;
			$q2=$pi*(1.0-$pj);
			if ($q1<$q2) {$dmax=$q1;} else {$dmax=$q2;}
		}
#		print "dprim = $dprim, hap = $hap, efr=$pij, refr=$hapf, dloc=$dloc, dmax=$dmax\n";
		$dprj += $pij * abs($dloc);
		if ($dmax>0) {
		$dppp = abs($dloc/$dmax);
		$dprim += $pij * $dppp;
		$dprim1 += $hapf * $dppp;
		$df++;
		$chiloc = ($znmb*$pij-$znmb*$hapf)*($znmb*$pij-$znmb*$hapf)/($znmb*$pij);
		$chisq += $chiloc;
#		print "$nm1 $key1 $cou1{$key1} $freq1{$key1} $nm2 $key2 $cou2{$key2} $freq2{$key2} $freqH{$hap} $hapf $pij $dloc $dppp $chiloc $dprim $chisq\n";
		} 
#		print "dprim = $dprim, hap = $hap, efr=$pij, refr=$hapf, dloc=$dloc\n";
	}
	}
	undef %cou1;undef %cou2;undef %couH;
	undef %freq1;undef %freq2;undef %freqH;
	return (1,$dprj,$dprim,$pval);
}
